import sys
import base64

from moban.plugins.jinja2.extensions import JinjaFilter


@JinjaFilter()
def base64encode(string):
    if sys.version_info[0] > 2:
        content = base64.b64encode(string.encode("utf-8"))
        content = content.decode("utf-8")
    else:
        content = base64.b64encode(string)
    return content



from moban.plugins.jinja2.extensions import JinjaFilter


@JinjaFilter()
def repr(string):
    if isinstance(string, list):
        return ["'{0}'".format(str(element)) for element in string]
    else:
        return "'{0}'".format(str(string))



from nose.tools import eq_

from moban.jinja2.filters.repr import repr as repr_function


def test_string():
    me = "abc"
    expected = repr_function(me)
    eq_(expected, "'abc'")


def test_list():
    me = [1, 2, 3]
    expected = repr_function(me)
    eq_(expected, ["'1'", "'2'", "'3'"])



from collections import defaultdict

import moban.constants as constants
import moban.exceptions as exceptions


class Strategy(object):
    DATA_FIRST = 1
    TEMPLATE_FIRST = 2

    def __init__(self, array_of_param_tuple):
        self.data_file_index = defaultdict(list)
        self.template_file_index = defaultdict(list)
        self.tuples = array_of_param_tuple

    def process(self):
        for target in self.tuples:
            _append_to_array_item_to_dictionary_key(
                self.data_file_index,
                target.data_file,
                (target.template_file, target.output),
            )
            _append_to_array_item_to_dictionary_key(
                self.template_file_index,
                target.template_file,
                (target.data_file, target.output),
            )

    def what_to_do(self):
        choice = Strategy.DATA_FIRST
        if self.data_file_index == {}:
            choice = Strategy.TEMPLATE_FIRST
        elif self.template_file_index != {}:
            data_files = len(self.data_file_index)
            template_files = len(self.template_file_index)
            if data_files > template_files:
                choice = Strategy.TEMPLATE_FIRST
        return choice


def _append_to_array_item_to_dictionary_key(adict, key, array_item):
    if array_item in adict[key]:
        raise exceptions.MobanfileGrammarException(
            constants.MESSAGE_SYNTAX_ERROR % (array_item, key)
        )
    else:
        adict[key].append(array_item)



import crayons

import moban.constants as constants

MESSAGE_TEMPLATING = "{0} {1} to {2}"
MESSAGE_UP_TO_DATE = "Everything is up to date!"
MESSAGE_NO_TEMPLATING = "No actions performed"
MESSAGE_REPORT = "{0} {1} out of {2} files."
MESSAGE_TEMPLATED_ALL = "{0} {1} files."
MESSAGE_PULLING_REPO = "Updating {0}..."
MESSAGE_CLONING_REPO = "Cloning {0}..."
MESSAGE_TEMPLATE_NOT_IN_MOBAN_FILE = "{0} is not defined in your moban file!"
MESSAGE_FILE_EXTENSION_NOT_NEEDED = "File extension is not required for ad-hoc\
 type"


def report_templating(
    action_in_present_continuous_tense, source_file, destination_file
):
    if destination_file:
        do_print(
            MESSAGE_TEMPLATING.format(
                action_in_present_continuous_tense,
                crayons.yellow(source_file),
                crayons.green(destination_file),
            )
        )
    else:
        do_print(
            f"{action_in_present_continuous_tense} {crayons.yellow(source_file)}"
        )


def report_no_action():
    do_print(crayons.yellow(MESSAGE_NO_TEMPLATING, bold=True))


def report_full_run(action_in_past_tense, file_count):
    figure = crayons.green(str(file_count), bold=True)
    message = MESSAGE_TEMPLATED_ALL.format(action_in_past_tense, figure)
    do_print(_format_single(message, file_count))


def report_partial_run(action_in_past_tense, file_count, total):
    figure = crayons.green(str(file_count), bold=True)
    total_figure = crayons.yellow(str(total), bold=True)
    message = MESSAGE_REPORT.format(action_in_past_tense, figure, total_figure)
    do_print(_format_single(message, total))


def report_error_message(message):
    do_print(crayons.white("Error: ", bold=True) + crayons.red(message))


def report_warning_message(message):
    do_print(crayons.white("Warning: ", bold=True) + crayons.yellow(message))


def report_info_message(message):
    do_print(crayons.white("Info: ") + crayons.green(message))


def report_up_to_date():
    do_print(crayons.green(MESSAGE_UP_TO_DATE, bold=True))


def convert_to_shell_exit_code(number_of_templated_files):
    return (
        constants.HAS_CHANGES
        if number_of_templated_files > 0
        else constants.NO_CHANGES
    )


def report_git_pull(repo):
    colored_repo = crayons.green(repo, bold=True)
    do_print(MESSAGE_PULLING_REPO.format(colored_repo))


def report_git_clone(repo):
    colored_repo = crayons.green(repo, bold=True)
    do_print(MESSAGE_CLONING_REPO.format(colored_repo))


def report_template_not_in_moban_file(template):
    message = MESSAGE_TEMPLATE_NOT_IN_MOBAN_FILE.format(template)
    report_warning_message(message)


def _format_single(message, count):
    if count == 1:
        return message.replace("files", "file")
    return message


def report_file_extension_not_needed():
    report_info_message(MESSAGE_FILE_EXTENSION_NOT_NEEDED)


def do_print(message):
    print(message)



import fs
from lml.plugin import PluginInfo

from moban import constants
from moban.core.mobanfile.store import STORE


@PluginInfo(
    constants.TEMPLATE_ENGINE_EXTENSION, tags=[constants.TEMPLATE_DELETE]
)
class DeleteEngine(object):
    """
    Does no templating but delete generated intermediate targets

    """

    ACTION_IN_PRESENT_CONTINUOUS_TENSE = "Deleting"
    ACTION_IN_PAST_TENSE = "Deleted"

    def __init__(self, template_fs, extensions=None):
        self.template_fs = template_fs

    def get_template(self, template_file):
        if template_file in STORE.intermediate_targets:
            with fs.open_fs(".") as the_fs:
                if the_fs.exists(template_file):
                    the_fs.remove(template_file)
                    return True
                else:
                    return False
        else:
            raise Exception(f"Cannot remove {template_file}")

    def get_template_from_string(self, string):
        raise NotImplementedError("Not sure what to do")

    def apply_template(self, template, *_):
        raise NotImplementedError("Not sure what to do")



from nose.tools import eq_

from moban.jinja2.filters.github import github_expand


def test_github_expand():
    inputs = [
        ["fix `#12`", "project", "organisation"],
        ["fix `#13` too", "project", "organisation"],
        ["`#14` works", "project", "organisation"],
        ["`PR#15` works", "project", "organisation"],
        ["`repo#PR#15` works", "project", "organisation"],
        ["`cool#16` works", "project", "organisation"],
        ["`microsoft#cool#PR#17` works", "project", "organisation"],
        ["this `cool##18`", "project", "organisation"],
        ["`goog#cool#19` works", "project", "organisation"],
        ["`twitter#cool#weird#19` works", "project", "organisation"],
        ["wont work", "project", "organisation"],
        ["`wont work either`", "prj", "org"],
        ["some `weird` `case` `#10` is", "prj", "org"],
    ]
    expectations = [
        "fix `#12 <https://github.com/organisation/project/issues/12>`_",
        "fix `#13 <https://github.com/organisation/project/issues/13>`_ too",
        "`#14 <https://github.com/organisation/project/issues/14>`_ works",
        "`#15 <https://github.com/organisation/project/pull/15>`_ works",
        "`repo#15 <https://github.com/organisation/repo/pull/15>`_ works",
        "`cool#16 <https://github.com/organisation/cool/issues/16>`_ works",
        "`cool#17 <https://github.com/microsoft/cool/pull/17>`_ works",
        "this `cool#18 <https://github.com/organisation/cool/issues/18>`_",
        "`goog#19 <https://github.com/organisation/goog/issues/19>`_ works",
        "`cool#19 <https://github.com/twitter/cool/issues/19>`_ works",
        "wont work",
        "`wont work either`",
        "some `weird` `case` `#10 <https://github.com/org/prj/issues/10>`_ is",
    ]
    for input_line, expect in zip(inputs, expectations):
        actual = github_expand(*input_line)
        eq_(actual, expect)



import logging
from collections import OrderedDict

from lml.plugin import PluginManager
from ruamel.yaml.comments import CommentedSeq

from moban import constants
from moban.externals import file_system

LOG = logging.getLogger()


class AnyDataLoader(PluginManager):
    def __init__(self):
        super(AnyDataLoader, self).__init__(constants.DATA_LOADER_EXTENSION)

    def get_data(self, file_name):
        file_extension = file_system.path_splitext(file_name)[1]
        file_type = file_extension
        if file_extension.startswith("."):
            file_type = file_type[1:]

        try:
            loader_function = self.load_me_now(file_type)
        except Exception:
            loader_function = self.load_me_now(constants.DEFAULT_DATA_TYPE)
        return loader_function(file_name)


LOADER = AnyDataLoader()


def load_data(base_dir, file_name):
    abs_file_path = search_file(base_dir, file_name)
    data = LOADER.get_data(abs_file_path)
    if data is not None:
        parent_data = OrderedDict()
        if constants.LABEL_OVERRIDES in data:
            overrides = data.pop(constants.LABEL_OVERRIDES)
            if not isinstance(overrides, list):
                overrides = [overrides]
            for parent_file in overrides:
                file_name, key = parent_file, None
                results = match_fs_url(parent_file)
                if results:
                    file_name, key = results
                elif ":" in parent_file and "://" not in parent_file:
                    file_name, key = parent_file.split(":")
                try:
                    child_data = load_data(base_dir, file_name)
                except IOError:
                    LOG.warn(f"failed to load {file_name} in {base_dir}")
                    child_data = {}
                if data:
                    if key:
                        child_data = OrderedDict({key: child_data[key]})
                    parent_data = merge(parent_data, child_data)
        if parent_data:
            return merge(data, parent_data)
        else:
            return data
    else:
        return None


def merge(left, right):
    """
    deep merge dictionary on the left with the one
    on the right.

    Fill in left dictionary with right one where
    the value of the key from the right one in
    the left one is missing or None.
    """
    if isinstance(left, dict) and isinstance(right, dict):
        for key, value in right.items():
            if key not in left:
                left[key] = value
            elif left[key] is None:
                left[key] = value
            else:
                left[key] = merge(left[key], value)
    else:
        both_list_alike = (
            isinstance(left, CommentedSeq) and isinstance(right, CommentedSeq)
        ) or (isinstance(left, list) and isinstance(right, list))
        if both_list_alike:
            left.extend(right)
    return left


def search_file(base_dir, file_name):
    the_file = file_name
    if not file_system.exists(the_file):
        if base_dir:
            file_under_base_dir = file_system.url_join(base_dir, the_file)
            if file_system.exists(file_under_base_dir):
                the_file = file_system.fs_url(file_under_base_dir)
            else:
                raise IOError(
                    constants.ERROR_DATA_FILE_NOT_FOUND % (file_name, the_file)
                )
        else:
            raise IOError(constants.ERROR_DATA_FILE_ABSENT % the_file)
    return the_file


def match_fs_url(file_name):
    import re

    results = re.match("(.*://.*):(.*)", file_name)
    if results:
        return (results.group(1), results.group(2))



class Store:
    def __init__(self):
        self.init()

    def add(self, target):
        self.targets.append(target)
        self.look_up_by_output[target.output] = target

    def init(self):
        self.targets = []
        self.look_up_by_output = {}
        self.intermediate_targets = []


STORE = Store()



import os
import re
import sys
import logging

from lml.utils import do_import

from moban import constants
from moban.externals import file_system

LOG = logging.getLogger(__name__)


def verify_the_existence_of_directories(dirs):
    if not isinstance(dirs, list):
        dirs = [dirs]

    dirs = [
        directory
        for directory in dirs
        if not (
            constants.DEFAULT_CONFIGURATION_DIRNAME in directory
            or constants.DEFAULT_TEMPLATE_DIRNAME in directory
        )
    ]
    if file_system.exists(constants.DEFAULT_CONFIGURATION_DIRNAME):
        dirs.append(constants.DEFAULT_CONFIGURATION_DIRNAME)

    if file_system.exists(constants.DEFAULT_TEMPLATE_DIRNAME):
        dirs.append(constants.DEFAULT_TEMPLATE_DIRNAME)
    return dirs


def handle_plugin_dirs(plugin_dirs):
    if plugin_dirs is None:
        return
    LOG.info("handling plugin dirs {}".format(",".join(plugin_dirs)))
    for plugin_dir in plugin_dirs:
        plugin_path = os.path.normcase(
            os.path.dirname(os.path.abspath(plugin_dir))
        )
        if plugin_path not in sys.path:
            sys.path.append(plugin_path)
        pysearchre = re.compile(".py$", re.IGNORECASE)
        pluginfiles = filter(pysearchre.search, os.listdir(plugin_dir))
        plugins = list(map(lambda fp: os.path.splitext(fp)[0], pluginfiles))
        for plugin in plugins:
            plugin_module = os.path.basename(plugin_dir) + "." + plugin
            do_import(plugin_module)



from moban.externals import file_system

# Template type
TEMPLATE_JINJA2 = "jinja2"
TEMPLATE_COPY = "copy"
TEMPLATE_DELETE = "delete"

# Configurations
PROGRAM_NAME = "moban"
PROGRAM_DESCRIPTION = (
    "Static text generator using any template, any data and any location."
)
DEFAULT_YAML_SUFFIX = ".yml"
# .moban.yml, default moban configuration file
DEFAULT_MOBAN_FILES = [
    ".%s%s" % (PROGRAM_NAME, DEFAULT_YAML_SUFFIX),
    ".%s%s" % (PROGRAM_NAME, ".yaml"),
]
DEFAULT_TEMPLATE_TYPE = TEMPLATE_JINJA2
DEFAULT_DATA_TYPE = "yaml"

# .moban.hashes
DEFAULT_MOBAN_CACHE_FILE = ".%s.hashes" % PROGRAM_NAME

# Command line options
LABEL_CONFIG = "configuration"
LABEL_CONFIG_DIR = "configuration_dir"
LABEL_PLUGIN_DIRS = "plugin_dir"
LABEL_TEMPLATE = "template"
POSITIONAL_LABEL_TEMPLATE = "template_in_string"
LABEL_TMPL_DIRS = "template_dir"
LABEL_PLUGIN_DIRS = "plugin_dir"
LABEL_EXIT_CODE = "exit-code"
LABEL_OUTPUT = "output"
LABEL_TEMPLATE_TYPE = "template_type"
LABEL_OVERRIDES = "overrides"
LABEL_MOBANFILE = "mobanfile"
LABEL_FORCE = "force"
LABEL_VERSION = "version"
LABEL_GROUP = "group"
LABEL_DEFINE = "define"
LABEL_EXTENSION = "extension"
CLI_DICT = "user_dict"
EXTENSION_DICT = "user_extensions"

DEFAULT_CONFIGURATION_DIRNAME = ".%s.cd" % PROGRAM_NAME
DEFAULT_TEMPLATE_DIRNAME = ".%s.td" % PROGRAM_NAME
DEFAULT_OPTIONS = {
    # .moban.cd, default configuration dir
    LABEL_CONFIG_DIR: file_system.path_join(
        ".", DEFAULT_CONFIGURATION_DIRNAME
    ),
    # .moban.td, default template dirs
    LABEL_TMPL_DIRS: [
        ".",
        file_system.path_join(".", DEFAULT_TEMPLATE_DIRNAME),
    ],
    # moban.output, default output file name
    LABEL_OUTPUT: "%s.output" % PROGRAM_NAME,
    # data.yml, default data input file
    LABEL_CONFIG: "data%s" % DEFAULT_YAML_SUFFIX,
    LABEL_TEMPLATE_TYPE: DEFAULT_TEMPLATE_TYPE,
}

# moban file version
MOBAN_VERSION = "moban_file_spec_version"
DEFAULT_MOBAN_VERSION = "1.0"
MOBAN_REPOS_DIR_NAME = "repos"
# moban file configuration
LABEL_REQUIRES = "requires"
LABEL_TARGETS = "targets"
LABEL_COPY = TEMPLATE_COPY
LABEL_EXTENSIONS = "extensions"
LABEL_ACTION = "action"
LABEL_SOURCE = "source"
LABEL_DEST = "destination"
LABEL_FORCE_TEMPLATE_TYPE = "force_template_type"
LABEL_TEMPLATE_TYPES = "template_types"
LABEL_VERBOSE = "verbose"
LABEL_MOBAN_ACTION_IN_PRESENT_CONTINUOUS_TENSE = "Mobanizing"
LABEL_MOBAN_ACTION_IN_PAST_TENSE = "Mobanized"

# error messages
ERROR_DATA_FILE_NOT_FOUND = "Both %s and %s does not exist"
ERROR_DATA_FILE_ABSENT = "File %s does not exist"

MESSAGE_SYNTAX_ERROR = "%s already exists in the target %s"
MESSAGE_DIR_NOT_EXIST = "%s does not exist"
MESSAGE_NO_THIRD_PARTY_ENGINE = "No such template support"
MESSAGE_FILE_VERSION_NOT_SUPPORTED = "moban file version '%s' is not supported"
MESSAGE_INVALID_GIT_URL = 'An invalid git url: "%s" in mobanfile'

# I/O messages
# Error handling
ERROR_INVALID_MOBAN_FILE = "%s is an invalid yaml file."
ERROR_NO_TARGETS = "No targets in %s"
ERROR_NO_TEMPLATE = "No template found"
ERROR_MALFORMED_YAML = "Empty value in %s at line %s"


# Shell messages
HAS_CHANGES = 1
ERROR = 2
NO_CHANGES = 0

# Require
DEFAULT_CLONE_DEPTH = 2
GIT_REQUIRE = "GIT"
GIT_HAS_SUBMODULE = "submodule"
GIT_URL = "url"
GIT_BRANCH = "branch"
GIT_REFERENCE = "reference"
PYPI_REQUIRE = "PYPI"
PYPI_PACKAGE_NAME = "name"
REQUIRE_TYPE = "type"

# Template types
TEMPLATE_TYPES_BASE_TYPE = "base_type"
TEMPLATE_TYPES_OPTIONS = "options"
TEMPLATE_TYPES_FILE_EXTENSIONS = "file_extensions"

# Extension
JINJA_FILTER_EXTENSION = "jinja_filter"
JINJA_TEST_EXTENSION = "jinja_test"
JINJA_GLOBALS_EXTENSION = "jinja_globals"

TEMPLATE_ENGINE_EXTENSION = "template_engine"
DATA_LOADER_EXTENSION = "data_loader"
LIBRARY_EXTENSION = "library"

MOBAN_EXTENSIONS = "^moban_.+$"
MOBAN_TEMPLATES = "^.+_mobans_pkg$"
MOBAN_ALL = "%s|%s" % (MOBAN_EXTENSIONS, MOBAN_TEMPLATES)


# deprecation message
MESSAGE_DEPRECATE_COPY_SINCE_0_4_0 = (
    "'%s:' is deprecated since 0.4.0! " + "Please use 'targets:' instead"
) % TEMPLATE_COPY
MESSAGE_DEPRECATE_MOBAN_NOTATION_SINCE_0_6_0 = (
    "File path notation is deprecated since 0.6.0! "
    + "Please use pypi:// or git:// notation instead"
)
MESSAGE_DEPRECATE_REQUIRES_SINCE_0_6_0 = (
    "'requires' is deprecated since 0.6.0! "
    + "Please use pypi:// or git:// notation instead"
)



import fs.path
from mock import patch
from nose.tools import eq_

from moban.core.mobanfile.templates import handle_template


class TestHandleTemplateFunction:
    def setUp(self):
        self.base_dir = [fs.path.join("tests", "fixtures")]

    def test_copy_files(self):
        results = list(
            handle_template("copier-test01.csv", "/tmp/test", self.base_dir)
        )
        expected = [("copier-test01.csv", "/tmp/test", "csv")]
        eq_(expected, results)

    @patch("moban.externals.reporter.report_error_message")
    def test_file_not_found(self, reporter):
        list(
            handle_template(
                "copier-test-not-found.csv", "/tmp/test", self.base_dir
            )
        )
        reporter.assert_called_with(
            "copier-test-not-found.csv cannot be found"
        )

    def test_listing_dir(self):
        test_dir = "/tmp/copy-a-directory"
        results = list(
            handle_template("copier-directory", test_dir, self.base_dir)
        )
        expected = [
            (
                "copier-directory/level1-file1",
                fs.path.join("/tmp/copy-a-directory", "level1-file1"),
                None,
            )
        ]
        eq_(expected, results)

    def test_listing_dir_recusively(self):
        test_dir = "/tmp/copy-a-directory"
        results = list(
            handle_template("copier-directory/**", test_dir, self.base_dir)
        )
        expected = [
            (
                fs.path.join("copier-directory", "copier-sample-dir", "file1"),
                fs.path.join(
                    "/tmp/copy-a-directory", "copier-sample-dir", "file1"
                ),
                None,
            ),
            (
                fs.path.join("copier-directory", "level1-file1"),
                fs.path.join("/tmp/copy-a-directory", "level1-file1"),
                None,
            ),
        ]
        eq_(
            sorted(results, key=lambda x: x[0]),
            sorted(expected, key=lambda x: x[0]),
        )

    @patch("moban.externals.reporter.report_error_message")
    def test_listing_dir_recusively_with_error(self, reporter):
        test_dir = "/tmp/copy-a-directory"
        list(
            handle_template(
                "copier-directory-does-not-exist/**", test_dir, self.base_dir
            )
        )
        eq_(reporter.call_count, 1)



"""
    moban
    ~~~~~~~~~~~~~~~~~~~

    Bring jinja2 to command line

    :copyright: (c) 2016-2020 by Onni Software Ltd.
    :license: MIT License, see LICENSE for more details

"""
import sys
import logging
import argparse
import logging.config
from io import StringIO
from collections import defaultdict

from ruamel.yaml import YAML

from moban import constants, exceptions
from moban.core import ENGINES, plugins, hashstore, mobanfile, data_loader
from moban._version import __version__
from moban.externals import reporter, file_system
from moban.core.utils import handle_plugin_dirs
from moban.program_options import OPTIONS

LOG = logging.getLogger()
LOG_LEVEL = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]


def main():
    parser = create_parser()
    options = vars(parser.parse_args())
    handle_verbose(options[constants.LABEL_VERBOSE])
    load_engine_factory_and_engines()  # Error: jinja2 if removed
    hashstore.HASH_STORE.IGNORE_CACHE_FILE = options[constants.LABEL_FORCE]
    options[constants.CLI_DICT] = handle_custom_variables(
        options.pop(constants.LABEL_DEFINE)
    )
    handle_custom_extensions(options.pop(constants.LABEL_EXTENSION))
    handle_plugin_dirs(options.pop(constants.LABEL_PLUGIN_DIRS))

    OPTIONS.update(options)
    moban_file = options[constants.LABEL_MOBANFILE]
    if moban_file is None:
        moban_file = find_default_moban_file()
    if moban_file:
        try:
            count = handle_moban_file(moban_file, options)
            moban_exit(options[constants.LABEL_EXIT_CODE], count)
        except (
            exceptions.DirectoryNotFound,
            exceptions.NoThirdPartyEngine,
            exceptions.MobanfileGrammarException,
            exceptions.UnsupportedPyFS2Protocol,
        ) as e:
            LOG.exception(e)
            reporter.report_error_message(str(e))
            moban_exit(options[constants.LABEL_EXIT_CODE], constants.ERROR)
    else:
        try:
            count = handle_command_line(options)
            moban_exit(options[constants.LABEL_EXIT_CODE], count)
        except (
            exceptions.NoTemplate,
            exceptions.UnsupportedPyFS2Protocol,
        ) as e:
            reporter.report_error_message(str(e))
            moban_exit(options[constants.LABEL_EXIT_CODE], constants.ERROR)


def moban_exit(exit_code_toggle_flag, exit_code):
    if exit_code_toggle_flag:
        if exit_code:
            sys.exit(exit_code)
    else:
        if exit_code == constants.ERROR:
            sys.exit(1)


def create_parser():
    parser = argparse.ArgumentParser(
        prog=constants.PROGRAM_NAME, description=constants.PROGRAM_DESCRIPTION
    )
    parser.add_argument(
        "-c", "--%s" % constants.LABEL_CONFIG, help="the data file"
    )
    parser.add_argument(
        "-t", "--%s" % constants.LABEL_TEMPLATE, help="the template file"
    )
    parser.add_argument(
        "-o", "--%s" % constants.LABEL_OUTPUT, help="the output file"
    )
    parser.add_argument(
        constants.POSITIONAL_LABEL_TEMPLATE,
        metavar="template",
        type=str,
        nargs="?",
        help="string templates",
    )
    advanced = parser.add_argument_group(
        "Advanced options", "For better control"
    )
    advanced.add_argument(
        "-td",
        f"--{constants.LABEL_TMPL_DIRS}",
        nargs="*",
        help="add more directories for template file lookup",
    )
    advanced.add_argument(
        "-cd",
        f"--{constants.LABEL_CONFIG_DIR}",
        help="the directory for configuration file lookup",
    )
    advanced.add_argument(
        "-pd",
        f"--{constants.LABEL_PLUGIN_DIRS}",
        nargs="*",
        help="add more directories for plugin lookup",
    )
    advanced.add_argument(
        "-m", "--%s" % constants.LABEL_MOBANFILE, help="custom moban file"
    )
    advanced.add_argument(
        "-g", "--%s" % constants.LABEL_GROUP, help="a subset of targets"
    )
    advanced.add_argument(
        f"--{constants.LABEL_TEMPLATE_TYPE.replace('_', '-')}",
        help="the template type, default is jinja2",
    )
    advanced.add_argument(
        "-d",
        f"--{constants.LABEL_DEFINE}",
        nargs="+",
        help=(
            "to supply additional or override predefined variables,"
            + " format: VAR=VALUEs"
        ),
    )
    advanced.add_argument(
        "-e",
        f"--{constants.LABEL_EXTENSION}",
        nargs="+",
        help="to to TEMPLATE_TYPE=EXTENSION_NAME",
    )
    advanced.add_argument(
        "-f",
        action="store_true",
        dest=constants.LABEL_FORCE,
        default=False,
        help="force moban to template all files despite of .moban.hashes",
    )
    developer = parser.add_argument_group(
        "Developer options", "For debugging and development"
    )
    developer.add_argument(
        f"--{constants.LABEL_EXIT_CODE}",
        action="store_true",
        dest=constants.LABEL_EXIT_CODE,
        default=False,
        help=(
            "by default, exist code 0 means no error, 1 means error occured. "
            + "It tells moban to change 1 for changes, 2 for error occured"
        ),
    )
    developer.add_argument(
        "-V",
        f"--{constants.LABEL_VERSION}",
        action="version",
        version="%(prog)s {v}".format(v=__version__),
    )
    developer.add_argument(
        "-v",
        action="count",
        dest=constants.LABEL_VERBOSE,
        default=0,
        help="show verbose, try -v, -vv, -vvv",
    )
    return parser


def handle_moban_file(moban_file, options):
    """
    act upon default moban file
    """
    moban_file_configurations = data_loader.load_data(None, moban_file)
    yaml = YAML(typ="rt")
    dumped_yaml = StringIO()
    yaml.dump(moban_file_configurations, dumped_yaml)
    LOG.info(dumped_yaml.getvalue())
    if moban_file_configurations is None:
        raise exceptions.MobanfileGrammarException(
            constants.ERROR_INVALID_MOBAN_FILE % moban_file
        )
    if (
        constants.LABEL_TARGETS not in moban_file_configurations
        and constants.LABEL_COPY not in moban_file_configurations
    ):
        raise exceptions.MobanfileGrammarException(
            constants.ERROR_NO_TARGETS % moban_file
        )
    check_none(moban_file_configurations, moban_file)
    version = moban_file_configurations.get(
        constants.MOBAN_VERSION, constants.DEFAULT_MOBAN_VERSION
    )
    if version == constants.DEFAULT_MOBAN_VERSION:
        mobanfile.handle_moban_file_v1(moban_file_configurations, options)
    else:
        raise exceptions.MobanfileGrammarException(
            constants.MESSAGE_FILE_VERSION_NOT_SUPPORTED % version
        )
    hashstore.HASH_STORE.save_hashes()


def check_none(data, moban_file):
    """
    check whether the yaml data has empty value such as:
    """
    if isinstance(data, dict):
        for k, v in data.items():
            if check_none(v, moban_file) is None:
                loc = data.lc.key(k)
                raise exceptions.MobanfileGrammarException(
                    constants.ERROR_MALFORMED_YAML
                    % (moban_file, loc[0] + 1)  # line number starts from 0
                )
    elif isinstance(data, list):
        for i, x in enumerate(data):
            if check_none(x, moban_file) is None:
                loc = data.lc.item(i)
                raise exceptions.MobanfileGrammarException(
                    constants.ERROR_MALFORMED_YAML % (moban_file, loc[0] + 1)
                )
    return data


def handle_command_line(options):
    """
    act upon command options
    """
    options = data_loader.merge(options, constants.DEFAULT_OPTIONS)
    engine = ENGINES.get_engine(
        options[constants.LABEL_TEMPLATE_TYPE],
        options[constants.LABEL_TMPL_DIRS],
        options[constants.LABEL_CONFIG_DIR],
    )
    if options[constants.LABEL_TEMPLATE] is None:
        content = options[constants.POSITIONAL_LABEL_TEMPLATE]
        if content is None:
            if not sys.stdin.isatty() and sys.platform != "win32":
                content = sys.stdin.read().strip()
        if content is None:
            raise exceptions.NoTemplate(constants.ERROR_NO_TEMPLATE)

        engine.render_string_to_file(
            content,
            options[constants.LABEL_CONFIG],
            options[constants.LABEL_OUTPUT],
        )
    else:
        engine.render_to_file(
            options[constants.LABEL_TEMPLATE],
            options[constants.LABEL_CONFIG],
            options[constants.LABEL_OUTPUT],
        )
    engine.report()
    hashstore.HASH_STORE.save_hashes()
    exit_code = reporter.convert_to_shell_exit_code(
        engine.number_of_templated_files()
    )
    return exit_code


def find_default_moban_file():
    for moban_file in constants.DEFAULT_MOBAN_FILES:
        if file_system.exists(moban_file):
            break
    else:
        moban_file = None
    return moban_file


def load_engine_factory_and_engines():
    plugins.make_sure_all_pkg_are_loaded()


def handle_custom_variables(list_of_definitions):
    custom_data = {}
    if list_of_definitions:
        for definition in list_of_definitions:
            key, value = definition.split("=")
            custom_data[key] = value

    return custom_data


def handle_custom_extensions(list_of_definitions):
    user_extensions = defaultdict(set)
    if list_of_definitions:
        for definition in list_of_definitions:
            key, value = definition.split("=")
            user_extensions[key].add(value)
    ENGINES.register_extensions(user_extensions)


def handle_verbose(verbose_level):
    if verbose_level > len(LOG_LEVEL):
        verbose_level = len(LOG_LEVEL) - 1
    level = LOG_LEVEL[verbose_level]
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        level=level,
    )



import logging
from collections import OrderedDict

from moban import core, constants
from moban.externals import reporter
from moban.core.utils import (
    handle_plugin_dirs,
    verify_the_existence_of_directories,
)
from moban.deprecated import handle_copy, handle_requires
from moban.core.data_loader import merge
from moban.core.moban_factory import expand_template_directories
from moban.core.mobanfile.targets import (
    parse_targets,
    extract_target,
    extract_group_targets,
)
from .store import STORE

LOG = logging.getLogger(__name__)


def handle_moban_file_v1(moban_file_configurations, command_line_options):
    LOG.info("handling moban file")
    merged_options = None

    targets = moban_file_configurations.get(constants.LABEL_TARGETS, [])
    if constants.LABEL_COPY in moban_file_configurations:
        legacy_copy_targets = handle_copy(
            merged_options, moban_file_configurations[constants.LABEL_COPY]
        )
        targets += legacy_copy_targets

    cli_target = extract_target(command_line_options)
    group_target = command_line_options.get(constants.LABEL_GROUP)
    if group_target:
        # will raise exception when group target not found
        targets = extract_group_targets(group_target, targets)

    if constants.LABEL_CONFIG in moban_file_configurations:
        merged_options = merge(
            command_line_options,
            moban_file_configurations[constants.LABEL_CONFIG],
        )
    merged_options = merge(command_line_options, constants.DEFAULT_OPTIONS)

    plugins_dirs = merged_options.get(constants.LABEL_PLUGIN_DIRS)
    if plugins_dirs:
        handle_plugin_dirs(plugins_dirs)

    # deprecated
    requires = moban_file_configurations.get(constants.LABEL_REQUIRES)
    if requires:
        handle_requires(requires)

    # call expand template directory always after handle require please
    # the penalty is: newly clone repos are not visible
    # one more note: verify_the_existence_of_directories will remove non-exist dirs
    merged_options[
        constants.LABEL_TMPL_DIRS
    ] = verify_the_existence_of_directories(
        list(
            expand_template_directories(
                merged_options[constants.LABEL_TMPL_DIRS]
            )
        )
    )

    extensions = moban_file_configurations.get(constants.LABEL_EXTENSIONS)
    if extensions:
        core.ENGINES.register_extensions(extensions)

    template_types = merged_options.get(constants.LABEL_TEMPLATE_TYPES)
    if template_types:
        core.ENGINES.register_options(template_types)

    if cli_target:
        number_of_templated_files = handle_targets(
            merged_options, [cli_target]
        )
    elif targets:
        number_of_templated_files = handle_targets(merged_options, targets)
    else:
        number_of_templated_files = 0

    exit_code = reporter.convert_to_shell_exit_code(number_of_templated_files)
    reporter.report_up_to_date()
    return exit_code


def handle_targets(merged_options, targets):
    LOG.info("handling targets")
    parse_targets(merged_options, targets)
    list_of_templating_parameters = STORE.targets
    jobs_for_each_engine = OrderedDict()

    for target in list_of_templating_parameters:
        forced_template_type = merged_options.get(
            constants.LABEL_FORCE_TEMPLATE_TYPE
        )
        if forced_template_type:
            target.set_template_type(forced_template_type)

        template_type = target.template_type
        primary_template_type = core.ENGINES.get_primary_key(template_type)
        if primary_template_type is None:
            primary_template_type = merged_options[
                constants.LABEL_TEMPLATE_TYPE
            ]
            target.set_template_type(primary_template_type)

        if primary_template_type not in jobs_for_each_engine:
            jobs_for_each_engine[primary_template_type] = []

        jobs_for_each_engine[primary_template_type].append(target)

    count = 0
    fall_out_targets = []
    for template_type in jobs_for_each_engine.keys():
        engine = core.ENGINES.get_engine(
            template_type,
            merged_options[constants.LABEL_TMPL_DIRS],
            merged_options[constants.LABEL_CONFIG_DIR],
        )
        engine.render_to_files(jobs_for_each_engine[template_type])
        engine.report()
        count = count + engine.number_of_templated_files()
        fall_out_targets += engine.fall_out_targets

    if fall_out_targets:
        copy_engine = core.ENGINES.get_engine(
            constants.TEMPLATE_COPY,
            merged_options[constants.LABEL_TMPL_DIRS],
            merged_options[constants.LABEL_CONFIG_DIR],
        )
        copy_engine.render_to_files(fall_out_targets)
        copy_engine.report()
        count = count + copy_engine.number_of_templated_files()
    return count



import os

import fs.path
from nose.tools import eq_

from moban.core.context import Context


def test_context():
    context = Context(fs.path.join("tests", "fixtures"))
    data = context.get_data("simple.yaml")
    eq_(data["simple"], "yaml")


def test_environ_variables():
    test_var = "TEST_ENVIRONMENT_VARIABLE"
    test_value = "am I found"
    os.environ[test_var] = test_value
    context = Context(fs.path.join("tests", "fixtures"))
    data = context.get_data("simple.yaml")
    eq_(data[test_var], test_value)


def test_json_data_overrides_environ_variables():
    test_var = "TEST_ENVIRONMENT_VARIABLE"
    test_value = "am I found"
    os.environ[test_var] = test_value
    context = Context(fs.path.join("tests", "fixtures"))
    data = context.get_data("simple.json")
    eq_(data[test_var], test_value)


def test_unknown_data_file():
    test_var = "TEST_ENVIRONMENT_VARIABLE"
    test_value = "am I found"
    os.environ[test_var] = test_value
    context = Context(fs.path.join("tests", "fixtures"))
    data = context.get_data("unknown.data")
    eq_(data[test_var], test_value)



import json

from lml.plugin import PluginInfo

from moban import constants
from moban.externals.file_system import open_file


@PluginInfo(constants.DATA_LOADER_EXTENSION, tags=["json"])
def open_json(file_name):
    """
    returns json contents as string
    """
    with open_file(file_name) as json_file:
        data = json.load(json_file)
        return data



import logging

from lml.plugin import PluginInfo

from moban import constants

LOG = logging.getLogger(__name__)


class ContentProcessor(PluginInfo):
    """
    @ContentProcessor('strip', 'Stripping', 'Stripped'):
    def strip(template_file: str) -> str:
        ret = template_file.strip()
        return ret
    """

    def __init__(self, action, action_continuing_tense, action_past_tense):
        super(ContentProcessor, self).__init__(
            constants.TEMPLATE_ENGINE_EXTENSION, tags=[action]
        )
        self.action = action
        self.action_continuing_tense = action_continuing_tense
        self.action_past_tense = action_continuing_tense

    def __call__(self, a_content_processor_function):
        continuing_tense = self.action_continuing_tense
        past_tense = self.action_past_tense

        class CustomEngine(object):
            ACTION_IN_PRESENT_CONTINUOUS_TENSE = continuing_tense
            ACTION_IN_PAST_TENSE = past_tense

            def __init__(self, template_fs, extensions=None):
                self.template_fs = template_fs

            def get_template(self, template_file):
                content = self.template_fs.readbytes(template_file)
                ret = a_content_processor_function(content)
                return ret

            def get_template_from_string(self, a_string):
                ret = a_content_processor_function(a_string)
                return ret

            def apply_template(self, template, *_):
                ret = a_content_processor_function(template)
                return ret

        super(ContentProcessor, self).__call__(CustomEngine)
        return a_content_processor_function



from moban.core.content_processor import ContentProcessor


@ContentProcessor("de-duplicate", "De-duplicating", "De-duplicated")
def de_duplicate(content: str) -> str:
    """
    Does no templating, works like 'copy'.

    """
    lines = content.split(b"\n")
    new_lines = []
    for line in lines:
        if line not in new_lines:
            new_lines.append(line)
    return b"\n".join(new_lines)



from moban.core.content_processor import ContentProcessor


@ContentProcessor("strip", "Stripping", "Stripped")
def strip(content: str) -> str:
    """Works like 'copy', but strip empty spaces before and after"""
    return content.strip()



import os
import sys
import filecmp

import fs.path
from mock import patch

from moban.main import main
from .utils import Docs


class TestRegression(Docs):
    def setUp(self):
        super(TestRegression, self).setUp()
        self.base_folder = fs.path.join("tests", "regression_tests")

    def test_coping_binary_file(self):
        folder = "regr-01-copy-binary-file"
        args = ["moban"]
        self._raw_moban(
            args,
            folder,
            fs.path.join("copy-source", "image.png"),
            "regression-test.png",
        )

    def test_default_copy_as_error_handling_behavior(self):
        folder = "regr-02-templating-failure-results-in-copy-action"
        args = ["moban"]
        self._raw_moban(
            args,
            folder,
            fs.path.join("copy-source", "image.png"),
            "regression-test.png",
        )

    def test_level_7(self):
        expected = "YWJj\n"

        folder = "level-7-plugin-dir-cli"
        self.run_moban(
            [
                "moban",
                "-td",
                "my-templates/",
                "-t",
                "filter.jj2",
                "-pd",
                "custom-jj2-plugin",
            ],
            folder,
            [("moban.output", expected)],
        )

    def test_level_7_b(self):
        expected = "1\n2\n3\n4\n5\n6\n8\n"

        folder = "level-7-b-template-engine-plugin"
        self.run_moban(
            [
                "moban",
                "--template-type",
                "de-duplicate",
                "-pd",
                "custom-plugin",
                "-t",
                "duplicated_content.txt",
            ],
            folder,
            [("moban.output", expected)],
        )

    def test_level_21_copy_templates_into_tars(self):
        expected = "test file\n"

        folder = "level-21-b-copy-templates-into-a-tar"
        long_url = (
            "tar://my.tar!/test-recursive-dir/sub_directory_is_copied"
            + "/because_star_star_is_specified.txt"
        )
        criterias = [
            ["tar://my.tar!/simple.file", expected],
            [
                "tar://my.tar!/target_without_template_type",
                "file extension will trigger copy engine\n",
            ],
            [
                "tar://my.tar!/target_in_short_form",
                (
                    "it is OK to have a short form, "
                    + "but the file to be 'copied' shall have 'copy' extension, "
                    + "so as to trigger ContentForwardEngine, 'copy' engine.\n"
                ),
            ],
            ["tar://my.tar!/test-dir/afile.txt", "dir for copying\n"],
            [long_url, "dest_directory: source_directory/**\n"],
        ]
        self.run_moban_with_fs(["moban"], folder, criterias)

    def test_level_21_copy_templates_from_tars(self):
        expected = "test file\n"

        folder = "level-21-c-copy-templates-from-a-tar"
        long_url = (
            "zip://my.zip!/test-recursive-dir/sub_directory_is_copied"
            + "/because_star_star_is_specified.txt"
        )
        criterias = [
            ["zip://my.zip!/simple.file", expected],
            [
                "zip://my.zip!/target_without_template_type",
                "file extension will trigger copy engine\n",
            ],
            [
                "zip://my.zip!/target_in_short_form",
                (
                    "it is OK to have a short form, "
                    + "but the file to be 'copied' shall have 'copy' extension, "
                    + "so as to trigger ContentForwardEngine, 'copy' engine.\n"
                ),
            ],
            ["zip://my.zip!/test-dir/afile.txt", "dir for copying\n"],
            [long_url, "dest_directory: source_directory/**\n"],
        ]
        self.run_moban_with_fs(["moban"], folder, criterias)

    def _raw_moban(self, args, folder, expected, output):
        base_dir = fs.path.join("tests", "regression_tests")
        os.chdir(fs.path.join(base_dir, folder))
        with patch.object(sys, "argv", args):
            main()
        status = filecmp.cmp(output, expected)
        os.unlink(output)
        assert status



from nose.tools import eq_

from moban.core.definitions import TemplateTarget
from moban.core.mobanfile.store import Store


def test_store():
    store = Store()
    output = "output"
    target = TemplateTarget("template_file", "data_file", output)
    store.add(target)
    eq_(target, store.look_up_by_output.get(output))



from lml.plugin import PluginManager

from moban import constants


class LibraryManager(PluginManager):
    def __init__(self):
        super(LibraryManager, self).__init__(constants.LIBRARY_EXTENSION)

    def resource_path_of(self, library_name):
        library = self.get_a_plugin(library_name)
        return library.resources_path


LIBRARIES = LibraryManager()



from moban.core.content_processor import ContentProcessor


@ContentProcessor("copy", "Copying", "Copied")
def copy(content: str) -> str:
    """
    Does no templating, works like 'copy'.

    Respects templating directories, for example: naughty.template
    could exist in any of template directires: dir1,
    dir2, dir3, and this engine will find it for you.  With conventional
    copy command, the source file path must be known.

    And this engine does not really touch the dest file but only read
    the source file. Everything else is taken care of by moban
    templating mechanism.
    """
    return content



from nose.tools import eq_

from moban.deprecated import GitRequire
from moban.core.definitions import TemplateTarget


def test_git_require_repr():
    require = GitRequire(git_url="http://github.com/some/repo")
    eq_("http://github.com/some/repo,None,False", repr(require))


def test_template_target_repr():
    require = TemplateTarget("template_file", "dat_file", "output")
    eq_("template_file,dat_file,output,jinja2", repr(require))


def test_template_target_output_suffix_change():
    require = TemplateTarget(
        "template_file", "dat_file", "output.copy", template_type="copy"
    )
    eq_("template_file,dat_file,output,copy", repr(require))


def test_template_target_output_suffix_updates_after_set():
    require = TemplateTarget(
        "template_file", "dat_file", "output.copy", template_type="copy"
    )
    require.set_template_type("jinja2")
    eq_("template_file,dat_file,output.copy,jinja2", repr(require))


def test_clone_params():
    require = GitRequire(git_url="http://github.com/some/repo")
    actual = require.clone_params()
    expected = {"single_branch": True, "depth": 2}
    eq_(expected, actual)


def test_branch_params():
    require = GitRequire(
        git_url="http://github.com/some/repo", branch="ghpages"
    )
    actual = require.clone_params()
    expected = {"single_branch": True, "branch": "ghpages", "depth": 2}
    eq_(expected, actual)



import os

import fs.path
from nose.tools import eq_

from moban.core import ENGINES
from moban.externals import file_system


class TestContentForwardEngine:
    def setUp(self):
        template_path = os.path.join("tests", "fixtures")
        fs = file_system.get_multi_fs([template_path])
        ContentForwardEngine = ENGINES.load_me_now("copy")
        self.engine = ContentForwardEngine(fs)

    def test_get_template(self):
        template_content = self.engine.get_template("copier-test01.csv")
        #  remove '\r' for windows
        eq_("test 01\n", template_content.decode("utf-8").replace("\r", ""))

    def test_encoding_of_template(self):
        template_content_ = self.engine.get_template("coala_color.svg")
        with open("tests/fixtures/coala_color.svg", "r") as expected:
            expected = expected.read()
        eq_(expected, template_content_.decode("utf-8").replace("\r", ""))

    def test_get_template_from_string(self):
        test_content = "simply forwarded"
        template_content = self.engine.get_template_from_string(test_content)
        eq_(test_content, template_content)

    def test_apply_template(self):
        test_content = "simply forwarded"
        template_content = self.engine.apply_template(test_content, "not used")
        eq_(test_content, template_content)


class TestCopyEncoding:
    def setUp(self):
        template_path = fs.path.join("tests", "fixtures")
        template_fs = file_system.get_multi_fs([template_path])
        ContentForwardEngine = ENGINES.load_me_now("copy")
        self.engine = ContentForwardEngine(template_fs)

    def test_encoding_of_template(self):
        template_content = self.engine.get_template("coala_color.svg")
        with open("tests/fixtures/coala_color.svg", "rb") as expected:
            expected = expected.read()
        eq_(expected, template_content)
        template_content = self.engine.get_template("non-unicode.char")
        with open("tests/fixtures/non-unicode.char", "rb") as expected:
            expected = expected.read()
        eq_(expected, template_content)



import re
import logging
from typing import Dict
from importlib import import_module

import fs.errors
from jinja2 import Template, Environment
from lml.loader import scan_plugins_regex
from lml.plugin import PluginInfo, PluginManager
from jinja2_fsloader import FSLoader
from jinja2.exceptions import TemplateNotFound, TemplateSyntaxError

from moban import constants, exceptions
from moban.externals import file_system

JINJA2_LIBRARIES = "^moban_jinja2_.+$"
JINJA2_EXTENSIONS = [
    "moban.plugins.jinja2.filters.repr",
    "moban.plugins.jinja2.filters.github",
    "moban.plugins.jinja2.filters.text",
    "moban.plugins.jinja2.tests.files",
]
JINJA2_THIRD_PARTY_EXTENSIONS = ["jinja2.ext.do", "jinja2.ext.loopcontrols"]
LOG = logging.getLogger(__name__)


class PluginMixin:
    def get_all(self):
        for name in self.registry.keys():
            # only the first matching one is returned
            the_filter = self.load_me_now(name)
            yield (name, the_filter)


class JinjaFilterManager(PluginManager, PluginMixin):
    def __init__(self):
        super(JinjaFilterManager, self).__init__(
            constants.JINJA_FILTER_EXTENSION
        )


class JinjaTestManager(PluginManager, PluginMixin):
    def __init__(self):
        super(JinjaTestManager, self).__init__(constants.JINJA_TEST_EXTENSION)


class JinjaGlobalsManager(PluginManager, PluginMixin):
    def __init__(self):
        super(JinjaGlobalsManager, self).__init__(
            constants.JINJA_GLOBALS_EXTENSION
        )


FILTERS = JinjaFilterManager()
TESTS = JinjaTestManager()
GLOBALS = JinjaGlobalsManager()


@PluginInfo(
    constants.TEMPLATE_ENGINE_EXTENSION, tags=["jinja2", "jinja", "jj2", "j2"]
)
class Engine(object):

    ACTION_IN_PRESENT_CONTINUOUS_TENSE = "Templating"
    ACTION_IN_PAST_TENSE = "Templated"

    def __init__(self, template_fs, options: Dict = None):
        """
        Contruct a jinja2 template engine

        an instance of fs.multifs.MultiFS will be given and straightaway it is given
        to jinja2.FSLoader.

        :param fs.multifs.MultiFS template_fs: a MultiFS instance or a FS instance
        :param dict options: a dictionary containing environmental parameters
        """
        LOG.debug("Jinja template engine started")
        load_jinja2_extensions()
        template_loader = FSLoader(template_fs)
        env_params = dict(
            loader=template_loader,
            keep_trailing_newline=True,
            trim_blocks=True,
            lstrip_blocks=True,
            extensions=[
                extension for extension in JINJA2_THIRD_PARTY_EXTENSIONS
            ],  # get a copy of this global variable
        )
        self.template_loader = template_loader

        if options:
            if "extensions" in options:
                extensions = options.pop("extensions")
                if is_extension_list_valid(extensions):
                    # because it is modified here
                    env_params["extensions"] += extensions
                    import_module_of_extension(extensions)
            env_params.update(options)
        self.jj2_environment = Environment(**env_params)
        for filter_name, filter_function in FILTERS.get_all():
            self.jj2_environment.filters[filter_name] = filter_function

        for test_name, test_function in TESTS.get_all():
            self.jj2_environment.tests[test_name] = test_function

        for global_name, dict_obj in GLOBALS.get_all():
            self.jj2_environment.globals[global_name] = dict_obj

    def get_template(self, template_file):
        """
        :param str template_file: the template file name that appeared in moban
                                  file. It could be a file name, or a relative
                                  file path with reference to template
                                  directories.
        :return: a jinja2 template

        For example:

        suppose your current working directory is: /User/moban-pro/ and your
        template folder list is: ['./my-template'], and the given template
        file equals to: 'templates/myfile.jj2', they as a group tells the
        template file exists at:
            '/User/moban-pro/my-template/templates/myfile.jj2'
        """
        try:
            template = self.jj2_environment.get_template(template_file)
            return template
        except TemplateNotFound:
            try:
                content = file_system.read_unicode(template_file)
                return self.jj2_environment.from_string(content)
            except fs.errors.ResourceNotFound:
                return self.jj2_environment.from_string(template_file)
        except (UnicodeDecodeError, TemplateSyntaxError) as e:
            raise exceptions.PassOn(str(e))

    def get_template_from_string(self, string):
        return Template(string)

    def apply_template(self, template, data, output):
        """
        It is not expected this function to write content to file system.
        Please just apply data inside the template and return utf-8
        content.

        :param template: a jinja2 template from :class:`.get_template`
        :param dict data: python data dictionary
        :param str output: output file name
        """
        template.globals["__target__"] = output
        template.globals["__template__"] = template.name
        rendered_content = template.render(**data)
        rendered_content = strip_off_trailing_new_lines(rendered_content)
        return rendered_content


def load_jinja2_extensions():
    scan_plugins_regex(JINJA2_LIBRARIES, "moban", None, JINJA2_EXTENSIONS)


def is_extension_list_valid(extensions):
    return (
        extensions is not None
        and isinstance(extensions, list)
        and len(extensions) > 0
    )


def import_module_of_extension(extensions):
    modules = set()
    if extensions:
        for extension in extensions:
            modules.add(extension.split(".")[0])
    for module in modules:
        import_module(module)


def strip_off_trailing_new_lines(content):
    return re.sub(r"(\n\s+)+$", r"\n", content)



import re

from moban.plugins.jinja2.extensions import JinjaFilter

GITHUB_REF_PATTERN = "`([^`]*?#[0-9]+)`"
ISSUE = "^.*?" + GITHUB_REF_PATTERN + ".*?$"
SAME_PROJ_FULL_ISSUE = "`#{3} <https://github.com/{0}/{1}/{2}/{3}>`_"
DIFF_PROJ_FULL_ISSUE = "`{1}#{3} <https://github.com/{0}/{1}/{2}/{3}>`_"
PULL_REQUEST = "PR"
PULL = "pull"
ISSUES = "issues"


@JinjaFilter()
def github_expand(line, name, organisation):
    result = re.match(ISSUE, line)
    if result:
        github_thing = result.group(1)
        tokens = github_thing.split("#")
        if len(tokens) == 4:
            if tokens[2] == PULL_REQUEST:
                tokens[2] = PULL
            else:
                tokens[2] = ISSUES
        elif len(tokens) == 3:
            if tokens[1] == PULL_REQUEST:
                tokens = [organisation, tokens[0], PULL, tokens[2]]
            else:
                tokens = [organisation, tokens[0], ISSUES, tokens[2]]
        elif len(tokens) == 2:
            if tokens[0] == PULL_REQUEST:
                tokens = [organisation, name, PULL] + tokens[1:]
            elif tokens[0] != "":
                tokens = [organisation, tokens[0], ISSUES] + tokens[1:]
            else:
                tokens = [organisation, name, ISSUES] + tokens[1:]
        if tokens[1] != name:
            reference = DIFF_PROJ_FULL_ISSUE.format(*tokens)
        else:
            reference = SAME_PROJ_FULL_ISSUE.format(*tokens)
        return re.sub(GITHUB_REF_PATTERN, reference, line)
    else:
        return line



import fs.path
from mock import patch
from nose.tools import eq_

from moban.core.definitions import TemplateTarget


@patch("moban.core.moban_factory.MobanEngine.render_to_files")
def test_handle_targets(fake_renderer):
    from moban.core.mobanfile import handle_targets

    TEMPLATE = "copier-test01.csv"
    OUTPUT = "output.csv"
    CONFIGURATION = "child.yaml"
    TEMPLATE_DIRS = [fs.path.join("tests", "fixtures")]
    DEFAULT_TEMPLATE_TYPE = "jinja2"

    options = dict(
        configuration=CONFIGURATION,
        template_type=DEFAULT_TEMPLATE_TYPE,
        template_dir=TEMPLATE_DIRS,
        configuration_dir=fs.path.join("tests", "fixtures"),
    )
    short_hand_targets = [{OUTPUT: TEMPLATE}]
    handle_targets(options, short_hand_targets)

    call_args = list(fake_renderer.call_args[0][0])
    eq_(
        call_args,
        [
            TemplateTarget(
                "copier-test01.csv",
                "child.yaml",
                "output.csv",
                template_type="jinja2",
            )
        ],
    )


@patch("moban.core.moban_factory.MobanEngine.render_to_files")
def test_handle_targets_sequence(fake_renderer):
    from moban.core.mobanfile import handle_targets
    from moban.core.mobanfile.store import STORE

    STORE.init()  # required to reset the store

    TEMPLATE1 = "a.template.jj2"
    OUTPUT1 = "filterme.handlebars"  # in the future, this could dynamic output
    OUTPUT2 = "filtered_output.txt"
    CONFIGURATION = "child.yaml"
    TEMPLATE_DIRS = [fs.path.join("tests", "fixtures", "mobanfile")]
    DEFAULT_TEMPLATE_TYPE = "jinja2"

    options = dict(
        configuration=CONFIGURATION,
        template_type=DEFAULT_TEMPLATE_TYPE,
        template_dir=TEMPLATE_DIRS,
        configuration_dir=fs.path.join("tests", "fixtures"),
    )
    short_hand_targets = [{OUTPUT1: TEMPLATE1}, {OUTPUT2: OUTPUT1}]
    handle_targets(options, short_hand_targets)

    call_args = list(fake_renderer.call_args_list)

    eq_(
        call_args[0][0][0][0],
        TemplateTarget(
            "a.template.jj2",
            "child.yaml",
            "filterme.handlebars",
            template_type="jj2",
        ),
    )
    eq_(
        call_args[1][0][0][0],
        TemplateTarget(
            "filterme.handlebars",
            "child.yaml",
            "filtered_output.txt",
            template_type="handlebars",
        ),
    )



import re

from moban.plugins.jinja2.extensions import JinjaFilter


@JinjaFilter()
def split_length(input_line, length):
    start = 0
    limit = length
    line = re.sub(r"\s+", " ", input_line)
    line_length = len(line)
    if line_length <= length:
        yield line
    else:
        while True:
            if " " in line[start : start + limit]:  # noqa
                # go back and find a space
                while limit > 0 and line[start + limit] != " ":
                    limit -= 1
            else:
                # full whole line is single unit
                # so go forward find a space
                while (start + limit) < len(line) and line[
                    start + limit
                ] != " ":
                    limit += 1

            yield line[start : start + limit]  # noqa
            start = start + limit + 1
            limit = length
            if len(line[start:]) < length or start + limit >= len(line):
                break

        yield line[start:]



import logging

from moban import constants

LOG = logging.getLogger(__name__)


class TemplateTarget(object):
    def __init__(
        self,
        template_file,
        data_file,
        output,
        template_type=constants.DEFAULT_TEMPLATE_TYPE,
    ):
        self.template_file = template_file
        self.data_file = data_file
        self.original_output = output
        self.template_type = template_type
        self.output = self.original_output

        self.set_template_type(template_type)
        LOG.info("create a target {}".format(self))

    def set_template_type(self, new_template_type):
        self.template_type = new_template_type
        if self.original_output.endswith(self.template_type):
            self.output = self.original_output.replace(
                "." + self.template_type, ""
            )
        else:
            self.output = self.original_output

    def __eq__(self, other):
        return (
            self.template_file == other.template_file
            and self.data_file == other.data_file
            and self.output == other.output
            and self.template_type == self.template_type
        )

    def __repr__(self):
        return "%s,%s,%s,%s" % (
            self.template_file,
            self.data_file,
            self.output,
            self.template_type,
        )



import os

from nose.tools import eq_

from .utils import Docs, custom_dedent


class TestTutorial(Docs):
    def test_level_1(self):
        expected = "world"
        folder = "level-1-jinja2-cli"
        self._moban(folder, expected)

    def test_level_1_custom_define(self):
        expected = "maailman"
        folder = "level-1-jinja2-cli"
        args = ["moban", "-d", "hello=maailman", "-t", "a.template"]
        self.run_moban(args, folder, [("moban.output", expected)])

    def test_level_2(self):
        expected = """
        ========header============

        world

        ========footer============
        """
        expected = custom_dedent(expected)
        folder = "level-2-template-inheritance"
        self._moban(folder, expected)

    def test_level_3(self):
        expected = """
        ========header============

        world

        shijie

        ========footer============
        """
        expected = custom_dedent(expected)
        folder = "level-3-data-override"
        self._moban(folder, expected)

    def test_level_4(self):
        expected = """
        ========header============

        world

        shijie

        ========footer============
        """
        expected = custom_dedent(expected)
        folder = "level-4-single-command"
        self.run_moban(["moban"], folder, [("a.output", expected)])

    def test_level_5(self):
        expected = """
        ========header============

        world

        shijie

        this demonstrates jinja2's include statement

        ========footer============
        """
        expected = custom_dedent(expected)
        folder = "level-5-custom-configuration"
        self.run_moban(["moban"], folder, [("a.output", expected)])

    def test_level_6(self):
        expected = """
        ========header============

        world2

        shijie

        this demonstrates jinja2's include statement

        ========footer============
        """
        expected = custom_dedent(expected)
        folder = "level-6-complex-configuration"
        self.run_moban(["moban"], folder, [("a.output2", expected)])

    def test_level_20(self):
        expected = """
        ========header============

        world2

        shijie

        this demonstrates jinja2's include statement

        ========footer============
        """
        expected = custom_dedent(expected)
        folder = "level-20-templates-configs-in-zip-or-tar"
        self.run_moban_with_fs(
            ["moban"], folder, [("zip://a.zip!/a.output2", expected)]
        )

    def test_level_7(self):
        expected = """
        Hello, you are in level 7 example

        Hello, you are not in level 7
        """
        expected = custom_dedent(expected)

        folder = "level-7-use-custom-jinja2-filter-test-n-global"
        self.run_moban(["moban"], folder, [("test.output", expected)])

    def test_level_8(self):
        expected = "it is a test\n"
        folder = "level-8-pass-a-folder-full-of-templates"
        check_file = os.path.join("templated-folder", "my")
        self.run_moban(["moban"], folder, [(check_file, expected)])

    def test_level_9(self):
        expected = "pypi-mobans: moban dependency as pypi package"
        folder = "level-9-moban-dependency-as-pypi-package"
        self.run_moban(["moban"], folder, [("test.txt", expected)])

    def test_level_24(self):
        expected = "pypi-mobans: files over http protocol"
        folder = "level-24-files-over-http"
        self.run_moban(["moban"], folder, [("test.txt", expected)])

    def test_level_9_deprecated(self):
        expected = "pypi-mobans: moban dependency as pypi package"
        folder = "deprecated-level-9-moban-dependency-as-pypi-package"
        self.run_moban(["moban"], folder, [("test.txt", expected)])

    def test_level_10(self):
        expected = "pypi-mobans: moban dependency as git repo"
        folder = "level-10-moban-dependency-as-git-repo"
        self.run_moban(["moban"], folder, [("test.txt", expected)])

    def test_level_10_deprecated(self):
        expected = "pypi-mobans: moban dependency as git repo"
        folder = "deprecated-level-10-moban-dependency-as-git-repo"
        self.run_moban(["moban"], folder, [("test.txt", expected)])

    def test_level_11(self):
        expected = "handlebars does not support inheritance\n"
        folder = "level-11-use-handlebars"
        self.run_moban(["moban"], folder, [("a.output", expected)])

    def test_level_12(self):
        expected_a = """
        world
        world
        world
        world
        """
        expected_b = """
        142
        42
        142
        """
        expected_a = custom_dedent(expected_a)
        expected_b = custom_dedent(expected_b)
        folder = "level-12-use-template-engine-extensions"
        self.run_moban(
            ["moban"],
            folder,
            [("a.output", expected_a), ("b.output", expected_b)],
        )

    def test_level_13_json(self):
        expected = """
        ========header============

        world from child.json

        shijie from parent.yaml

        ========footer============
        """
        expected = custom_dedent(expected)
        folder = "level-13-any-data-override-any-data"
        commands = ["moban", "-c", "child.json", "-t", "a.template"]
        self.run_moban(commands, folder, [("moban.output", expected)])

    def test_level_13_yaml(self):
        expected = """
        ========header============

        world from child.yaml

        shijie from parent.json

        ========footer============
        """
        expected = custom_dedent(expected)
        folder = "level-13-any-data-override-any-data"
        commands = ["moban", "-c", "child.yaml", "-t", "a.template"]
        self.run_moban(commands, folder, [("moban.output", expected)])

    def test_level_14_custom(self):
        expected = """
        ========header============

        world from child.cusom

        shijie from parent.json

        ========footer============
        """
        expected = custom_dedent(expected)
        folder = "level-14-custom-data-loader"
        commands = ["moban"]
        self.run_moban(commands, folder, [("a.output", expected)])

    def test_level_15_copy_templates_as_target(self):
        expected = "test file\n"

        folder = "level-15-copy-templates-as-target"
        assertions = [
            ("simple.file", expected),
            (
                "target_without_template_type",
                "file extension will trigger copy engine\n",
            ),
            (
                "target_in_short_form",
                (
                    "it is OK to have a short form, "
                    + "but the file to be 'copied' shall have 'copy' extension, "
                    + "so as to trigger ContentForwardEngine, 'copy' engine.\n"
                ),
            ),
            (
                "output_is_copied.same_file_extension",
                "it is implicit copy as well",
            ),
        ]
        self.run_moban(["moban"], folder, assertions)

    def test_level_21_copy_templates_into_zips(self):
        expected = "test file\n"

        folder = "level-21-copy-templates-into-an-alien-file-system"
        long_url = (
            "zip://my.zip!/test-recursive-dir/sub_directory_is_copied"
            + "/because_star_star_is_specified.txt"
        )
        criterias = [
            ["zip://my.zip!/simple.file", expected],
            [
                "zip://my.zip!/target_without_template_type",
                "file extension will trigger copy engine\n",
            ],
            [
                "zip://my.zip!/target_in_short_form",
                (
                    "it is OK to have a short form, "
                    + "but the file to be 'copied' shall have 'copy' extension, "
                    + "so as to trigger ContentForwardEngine, 'copy' engine.\n"
                ),
            ],
            ["zip://my.zip!/test-dir/afile.txt", "dir for copying\n"],
            [long_url, "dest_directory: source_directory/**\n"],
        ]
        self.run_moban_with_fs(["moban"], folder, criterias)

    def test_level_16_group_targets_using_template_type(self):
        expected = "test file\n"

        folder = "level-16-group-targets-using-template-type"
        self.run_moban(["moban"], folder, [("simple.file", expected)])

    def test_level_17_force_template_type_from_moban_file(self):
        expected = "test file\n"

        folder = "level-17-force-template-type-from-moban-file"
        self.run_moban(["moban"], folder, [("simple.file", expected)])

    def test_level_18_user_defined_template_types(self):
        from datetime import datetime

        expected = "{date}\n".format(date=datetime.now().strftime("%Y-%m-%d"))

        folder = "level-18-user-defined-template-types"
        self.run_moban(
            ["moban"],
            folder,
            [("a.output", expected), ("b.output", "shijie\n")],
        )

    def test_level_19_without_group_target(self):
        expected = "test file\n"

        folder = "level-19-moban-a-sub-group-in-targets"
        assertions = [
            ("simple.file", expected),
            ("a.output", "I will not be selected in level 19\n"),
        ]
        self.run_moban(["moban"], folder, assertions)

    def test_level_19_with_group_target(self):
        expected = "test file\n"

        folder = "level-19-moban-a-sub-group-in-targets"
        self.run_moban(
            ["moban", "-g", "copy"], folder, [("simple.file", expected)]
        )
        # make sure only copy target is executed
        eq_(False, os.path.exists("a.output"))

    def test_level_22_intermediate_targets(self):
        expected = "a world\n"

        folder = "level-22-intermediate-targets"
        self.run_moban(["moban"], folder, [("final", expected)])
        assert os.path.exists("intermediate.jj2")

    def test_level_25_delete_intermediate_targets(self):
        expected = "a world\n"

        folder = "level-25-delete-intermediate-targets"
        self.run_moban(["moban"], folder, [("final", expected)])
        assert not os.path.exists("intermediate.jj2")
        assert not os.path.exists("intermediate2.jj2")
        assert not os.path.exists("intermediate3.jj2")

    def test_level_26_strip_intermediate_targets(self):
        expected = "a world"

        folder = "level-26-strip-rendered-content"
        self.run_moban(["moban"], folder, [("final", expected)])
        assert not os.path.exists("intermediate.strip")

    def test_level_23_inherit_parent_moban_file(self):
        folder = "level-23-inherit-organisational-moban-file"
        self.run_moban(
            ["moban"],
            folder,
            [("output_a", "I am template a"), ("output_b", "I am template b")],
        )

    def test_misc_1(self):
        expected = "test file\n"

        folder = "misc-1-copying-templates"
        self.run_moban(["moban"], folder, [("simple.file", expected)])

    def _moban(self, folder, expected):
        args = ["moban", "-c", "data.yml", "-t", "a.template"]
        self.run_moban(args, folder, [("moban.output", expected)])



import os
import sys
import stat
from shutil import rmtree

from mock import patch
from nose import SkipTest
from nose.tools import eq_, raises

from moban.externals import file_system
from moban.exceptions import FileNotFound, UnsupportedPyFS2Protocol

LOCAL_FOLDER = "tests/fixtures"
LOCAL_FILE = LOCAL_FOLDER + "/a.jj2"
ZIP_FILE = "zip://tests/fixtures/file_system/template-sources.zip"
TAR_FILE = "tar://tests/fixtures/file_system/template-sources.tar"

FILE = "file-in-template-sources-folder.txt"

ZIP_URL = ZIP_FILE + "!/" + FILE
TAR_URL = TAR_FILE + "!/" + FILE

TEST_SPECS = [LOCAL_FILE, ZIP_URL, TAR_URL]


TEST_FS_SPECS = [LOCAL_FOLDER, ZIP_FILE, TAR_FILE]


def test_open_file():
    for url in TEST_SPECS:
        with file_system.open_file(url):
            pass


def test_open_fs():
    for url in TEST_FS_SPECS:
        with file_system.open_fs(url):
            pass


TEST_FILE_CONTENT_SPECS = [
    [LOCAL_FILE, "{{key}} {{pass}}"],
    [ZIP_URL, "test file\n"],
    [TAR_URL, "test file\n"],
]


def test_read_unicode():
    for url, expected in TEST_FILE_CONTENT_SPECS:
        content = file_system.read_unicode(url)
        eq_(content, expected)


TEST_FILE_CONTENT_SPECS_BINARY = [
    [LOCAL_FILE, b"{{key}} {{pass}}"],
    [ZIP_URL, b"test file\n"],
    [TAR_URL, b"test file\n"],
]


def test_read_binary():
    for url, expected in TEST_FILE_CONTENT_SPECS_BINARY:
        content = file_system.read_binary(url)
        eq_(content, expected)


TEST_WRITE_BYTES_SPEC = [
    ["test.binary", b"abc"],
    ["zip://test.zip!/test.binary", b"abc"],
    ["tar://test.tar!/test.binary", b"abc"],
]


def test_write_bytes():
    for url, content in TEST_WRITE_BYTES_SPEC:
        file_system.write_bytes(url, content)

    for url, expected in TEST_WRITE_BYTES_SPEC:
        content = file_system.read_bytes(url)
        eq_(content, expected)

    for file_name in ["test.binary", "test.zip", "test.tar"]:
        os.unlink(file_name)


TEST_DIR_SPEC = [
    [LOCAL_FOLDER, True],
    [ZIP_FILE + "!/dir-for-copying", True],
    [TAR_FILE + "!/dir-for-copying", True],
    [ZIP_URL, False],
    [TAR_URL, False],
    [LOCAL_FILE, False],
]


def test_is_dir():
    for url, expected in TEST_DIR_SPEC:
        status = file_system.is_dir(url)
        eq_(status, expected)


def test_is_file():
    for url, is_dir in TEST_DIR_SPEC:
        status = file_system.is_file(url)
        expected = not is_dir
        eq_(status, expected)


TEST_URL_EXITENCE_SPEC = [
    [LOCAL_FOLDER, True],
    [ZIP_FILE + "!/dir-for-copying", True],
    [TAR_FILE + "!/dir-for-copying", True],
    [ZIP_URL, True],
    [TAR_URL, True],
    [LOCAL_FILE, True],
    # ['zip://abc.zip', False],
    # ['tar://abc.tar', False], bug with fs.zipfs. raise it later
    ["abcx", False],
]


def test_exists():
    for url, expected in TEST_URL_EXITENCE_SPEC:
        status = file_system.exists(url)
        eq_(status, expected)


@raises(UnsupportedPyFS2Protocol)
def test_exists_raise_exception():
    file_system.exists("git2://protocol/abc")


@raises(UnsupportedPyFS2Protocol)
def test_is_file_raise_exception():
    file_system.is_file("git2://protocol/abc")


TEST_LIST_DIR_SPEC = [
    [
        LOCAL_FOLDER + "/file_system",
        ["template-sources.zip", "template-sources.tar"],
    ],
    [
        ZIP_FILE + "!/dir-for-copying",
        ["afile.txt", "sub_directory_is_not_copied"],
    ],
    [
        TAR_FILE + "!/dir-for-copying",
        ["afile.txt", "sub_directory_is_not_copied"],
    ],
]


def test_list_dir():
    for url, expected in TEST_LIST_DIR_SPEC:
        file_list = sorted(list(file_system.list_dir(url)))
        eq_(file_list, sorted(expected))


TEST_FILE_PATH = [
    [
        LOCAL_FOLDER + "/file_system",
        os.path.normpath(
            os.path.join(os.getcwd(), "tests/fixtures/file_system")
        ),
    ]
]


def test_abspath():
    for path, expected in TEST_FILE_PATH:
        url = file_system.abspath(path)
        eq_(url, expected)


TEST_FILE_URL = [
    [
        LOCAL_FOLDER + "/file_system",
        "osfs://"
        + os.path.normpath(
            os.path.join(os.getcwd(), "tests/fixtures/file_system")
        ),
    ]
]


def test_fs_url():
    for path, expected in TEST_FILE_URL:
        url = file_system.fs_url(path)
        eq_(url, expected.replace("\\", "/"))


URL_JOIN_TEST_FIXTURES = [
    ["parent", "child", "parent/child"],
    ["zip://test.zip", "file", "zip://test.zip!/file"],
    ["/root", "path", "/root/path"],
]


def test_url_join():
    for parent, child, expected_path in URL_JOIN_TEST_FIXTURES:
        actual = file_system.url_join(parent, child)
        eq_(actual, expected_path)


def create_file(test_file, permission):
    with open(test_file, "w") as f:
        f.write("test")

    os.chmod(test_file, permission)


def test_file_permission_copy():
    if sys.platform == "win32":
        raise SkipTest("No actual chmod on windows")
    test_source = "test_file_permission_copy1"
    test_dest = "test_file_permission_copy2"
    create_file(test_source, 0o755)
    create_file(test_dest, 0o646)
    file_system.file_permissions_copy(test_source, test_dest)
    eq_(
        stat.S_IMODE(os.lstat(test_source).st_mode),
        stat.S_IMODE(os.lstat(test_dest).st_mode),
    )
    os.unlink(test_source)
    os.unlink(test_dest)


def file_permissions_disabled_on_windows():
    if sys.platform == "win32":
        permissions = file_system.file_permissions("abc")
        eq_("no-permission-support", permissions)
    else:
        raise SkipTest("No test required")


@raises(FileNotFound)
def test_file_permissions_file_not_found():
    file_system.file_permissions("I does not exist")


def test_file_permission_copy_symlink():
    if sys.platform == "win32":
        raise SkipTest("No symlink on windows")
    test_source = "test_file_permission_copy1"
    test_dest = "test_file_permission_copy2"
    test_symlink = "test_file_permission_symlink"
    create_file(test_source, 0o046)
    os.symlink(test_source, test_symlink)
    create_file(test_dest, 0o646)
    file_system.file_permissions_copy(test_source, test_dest)
    eq_(
        stat.S_IMODE(os.lstat(test_source).st_mode),
        stat.S_IMODE(os.lstat(test_dest).st_mode),
    )
    os.unlink(test_source)
    os.unlink(test_dest)
    os.unlink(test_symlink)


def test_mkdir_p():
    test_path = "a/b/c/d"
    file_system.mkdir_p(test_path)
    assert os.path.exists(test_path)
    rmtree(test_path)


@patch("subprocess.check_call")
def test_pip_install(fake_check_all):
    import sys
    from moban.deprecated import pip_install

    pip_install(["package1", "package2"])
    fake_check_all.assert_called_with(
        [sys.executable, "-m", "pip", "install", "package1 package2"],
        stderr=-3,
        stdout=-3,
    )



import os
import sys
from shutil import copyfile

from mock import MagicMock, patch
from nose import SkipTest
from nose.tools import eq_, raises, assert_raises

from moban.core.definitions import TemplateTarget

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class TestCustomOptions:
    def setUp(self):
        self.config_file = "config.yaml"
        with open(self.config_file, "w") as f:
            f.write("hello: world")
        self.patcher1 = patch(
            "moban.core.utils.verify_the_existence_of_directories"
        )
        self.patcher1.start()

    @patch("moban.externals.file_system.abspath")
    @patch("moban.core.moban_factory.MobanEngine.render_to_file")
    def test_custom_options(self, fake_template_doer, fake_abspath):
        test_args = [
            "moban",
            "-c",
            self.config_file,
            "-cd",
            ".",
            "-td",
            ".",
            "-t",
            "a.jj2",
        ]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            fake_template_doer.assert_called_with(
                "a.jj2", "config.yaml", "moban.output"
            )

    @patch("moban.core.moban_factory.MobanEngine.render_to_file")
    def test_minimal_options(self, fake_template_doer):
        test_args = ["moban", "-c", self.config_file, "-t", "a.jj2"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            fake_template_doer.assert_called_with(
                "a.jj2", "config.yaml", "moban.output"
            )

    @raises(SystemExit)
    def test_missing_template(self):
        test_args = ["moban", "-c", self.config_file]
        fake_stdin = MagicMock(isatty=MagicMock(return_value=True))
        with patch.object(sys, "stdin", fake_stdin):
            with patch.object(sys, "argv", test_args):
                from moban.main import main

                main()

    def tearDown(self):
        self.patcher1.stop()
        os.unlink(self.config_file)


class TestOptions:
    def setUp(self):
        self.config_file = "data.yml"
        with open(self.config_file, "w") as f:
            f.write("hello: world")
        self.patcher1 = patch(
            "moban.core.utils.verify_the_existence_of_directories"
        )
        self.patcher1.start()

    @patch("moban.core.moban_factory.MobanEngine.render_to_file")
    def test_default_options(self, fake_template_doer):
        test_args = ["moban", "-t", "a.jj2"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            fake_template_doer.assert_called_with(
                "a.jj2", "data.yml", "moban.output"
            )

    @patch("moban.core.moban_factory.MobanEngine.render_string_to_file")
    def test_string_template(self, fake_template_doer):
        string_template = "{{HELLO}}"
        test_args = ["moban", string_template]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            fake_template_doer.assert_called_with(
                string_template, "data.yml", "moban.output"
            )

    @raises(SystemExit)
    def test_no_argments(self):
        test_args = ["moban"]
        fake_stdin = MagicMock(isatty=MagicMock(return_value=True))
        with patch.object(sys, "stdin", fake_stdin):
            with patch.object(sys, "argv", test_args):
                from moban.main import main

                main()

    def tearDown(self):
        self.patcher1.stop()
        os.unlink(self.config_file)


class TestNoOptions:
    def setUp(self):
        self.config_file = ".moban.yml"
        copyfile(
            os.path.join("tests", "fixtures", self.config_file),
            self.config_file,
        )
        self.data_file = "data.yaml"
        with open(self.data_file, "w") as f:
            f.write("hello: world")
        self.patcher1 = patch(
            "moban.core.utils.verify_the_existence_of_directories"
        )
        self.patcher1.start()

    @patch("moban.core.moban_factory.MobanEngine.render_to_files")
    def test_single_command(self, fake_template_doer):
        test_args = ["moban"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            call_args = list(fake_template_doer.call_args[0][0])
            eq_(
                call_args,
                [
                    TemplateTarget(
                        "README.rst.jj2", "data.yaml", "README.rst"
                    ),
                    TemplateTarget("setup.py.jj2", "data.yaml", "setup.py"),
                ],
            )

    @raises(Exception)
    @patch("moban.core.moban_factory.MobanEngine.render_to_files")
    def test_single_command_with_missing_output(self, fake_template_doer):
        test_args = ["moban", "-t", "README.rst.jj2"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()

    @patch("moban.core.moban_factory.MobanEngine.render_to_files")
    def test_single_command_with_a_few_options(self, fake_template_doer):
        test_args = ["moban", "-t", "README.rst.jj2", "-o", "xyz.output"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()

            call_args = list(fake_template_doer.call_args[0][0])
            eq_(
                call_args,
                [TemplateTarget("README.rst.jj2", "data.yaml", "xyz.output")],
            )

    @patch("moban.core.moban_factory.MobanEngine.render_to_files")
    def test_single_command_with_options(self, fake_template_doer):
        test_args = [
            "moban",
            "-t",
            "README.rst.jj2",
            "-c",
            "new.yml",
            "-o",
            "xyz.output",
        ]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            call_args = list(fake_template_doer.call_args[0][0])
            eq_(
                call_args,
                [TemplateTarget("README.rst.jj2", "new.yml", "xyz.output")],
            )

    @raises(Exception)
    def test_single_command_without_output_option(self):
        test_args = ["moban", "-t", "abc.jj2"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()

    def tearDown(self):
        os.unlink(self.config_file)
        os.unlink(self.data_file)
        self.patcher1.stop()


class TestNoOptions2:
    def setUp(self):
        self.config_file = ".moban.yml"
        copyfile(
            os.path.join("tests", "fixtures", self.config_file),
            self.config_file,
        )
        self.data_file = "data.yaml"
        with open(self.data_file, "w") as f:
            f.write("hello: world")
        self.patcher1 = patch(
            "moban.core.utils.verify_the_existence_of_directories"
        )
        self.patcher1.start()

    @patch("moban.core.moban_factory.MobanEngine.render_to_files")
    def test_single_command(self, fake_template_doer):
        test_args = ["moban"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            call_args = list(fake_template_doer.call_args[0][0])
            eq_(
                call_args,
                [
                    TemplateTarget(
                        "README.rst.jj2", "data.yaml", "README.rst"
                    ),
                    TemplateTarget("setup.py.jj2", "data.yaml", "setup.py"),
                ],
            )

    def tearDown(self):
        self.patcher1.stop()
        os.unlink(self.config_file)
        os.unlink(self.data_file)


class TestCustomMobanFile:
    def setUp(self):
        self.config_file = "custom-moban.txt"
        copyfile(
            os.path.join("tests", "fixtures", ".moban.yml"), self.config_file
        )
        self.data_file = "data.yaml"
        with open(self.data_file, "w") as f:
            f.write("hello: world")
        self.patcher1 = patch(
            "moban.core.utils.verify_the_existence_of_directories"
        )
        self.patcher1.start()

    @patch("moban.core.moban_factory.MobanEngine.render_to_files")
    def test_single_command(self, fake_template_doer):
        test_args = ["moban", "-m", self.config_file]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            call_args = list(fake_template_doer.call_args[0][0])
            eq_(
                call_args,
                [
                    TemplateTarget(
                        "README.rst.jj2", "data.yaml", "README.rst"
                    ),
                    TemplateTarget("setup.py.jj2", "data.yaml", "setup.py"),
                ],
            )

    def tearDown(self):
        self.patcher1.stop()
        os.unlink(self.config_file)
        os.unlink(self.data_file)


class TestTemplateOption:
    def setUp(self):
        self.config_file = "custom-moban.txt"
        copyfile(
            os.path.join("tests", "fixtures", ".moban.yml"), self.config_file
        )
        self.patcher1 = patch(
            "moban.core.utils.verify_the_existence_of_directories"
        )
        self.patcher1.start()

    @patch("moban.core.moban_factory.MobanEngine.render_to_file")
    def test_template_option_override_moban_file(self, fake_template_doer):
        test_args = ["moban", "-t", "setup.py.jj2"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            fake_template_doer.assert_called_with(
                "setup.py.jj2", "data.yml", "moban.output"
            )

    @patch("moban.core.moban_factory.MobanEngine.render_to_file")
    def test_template_option_not_in_moban_file(self, fake_template_doer):
        test_args = ["moban", "-t", "foo.jj2"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            fake_template_doer.assert_called_with(
                "foo.jj2", "data.yml", "moban.output"
            )

    def tearDown(self):
        self.patcher1.stop()
        os.unlink(self.config_file)


@patch("moban.core.utils.verify_the_existence_of_directories")
def test_duplicated_targets_in_moban_file(fake_verify):
    config_file = "duplicated.moban.yml"
    copyfile(os.path.join("tests", "fixtures", config_file), ".moban.yml")
    test_args = ["moban"]
    with patch.object(sys, "argv", test_args):
        from moban.main import main

        assert_raises(SystemExit, main)
    os.unlink(".moban.yml")


class TestInvalidMobanFile:
    def setUp(self):
        self.config_file = ".moban.yml"

    @raises(SystemExit)
    @patch("moban.core.moban_factory.MobanEngine.render_to_files")
    def test_no_configuration(self, fake_template_doer):
        with open(self.config_file, "w") as f:
            f.write("")
        test_args = ["moban"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()

    @raises(SystemExit)
    @patch("moban.core.moban_factory.MobanEngine.render_to_files")
    def test_no_configuration_2(self, fake_template_doer):
        with open(self.config_file, "w") as f:
            f.write("not: related")
        test_args = ["moban"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()

    @raises(SystemExit)
    @patch("moban.core.moban_factory.MobanEngine.render_to_files")
    def test_no_targets(self, fake_template_doer):
        with open(self.config_file, "w") as f:
            f.write("configuration: test")
        test_args = ["moban"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()

    def tearDown(self):
        os.unlink(self.config_file)


class TestComplexOptions:
    def setUp(self):
        self.config_file = ".moban.yml"
        copyfile(
            os.path.join("tests", "fixtures", ".moban-2.yml"), self.config_file
        )
        self.data_file = "data.yaml"
        with open(self.data_file, "w") as f:
            f.write("hello: world")

    @patch(
        "moban.core.utils.verify_the_existence_of_directories",
        return_value=".",
    )
    def test_single_command(self, _):
        test_args = ["moban"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            with patch(
                "moban.core.moban_factory.MobanEngine.render_to_files"
            ) as fake:
                main()
                call_args = list(fake.call_args[0][0])
                eq_(
                    call_args,
                    [
                        TemplateTarget(
                            "README.rst.jj2", "custom-data.yaml", "README.rst"
                        ),
                        TemplateTarget("setup.py.jj2", "data.yml", "setup.py"),
                    ],
                )

    def tearDown(self):
        os.unlink(self.config_file)
        os.unlink(self.data_file)


class TestTemplateTypeOption:
    def setUp(self):
        self.config_file = "data.yml"
        with open(self.config_file, "w") as f:
            f.write("hello: world")

    @patch("moban.core.moban_factory.MobanEngine.render_to_file")
    def test_mako_option(self, fake_template_doer):
        test_args = ["moban", "-t", "a.mako"]
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            fake_template_doer.assert_called_with(
                "a.mako", "data.yml", "moban.output"
            )

    def tearDown(self):
        os.unlink(self.config_file)


@raises(SystemExit)
def test_version_option():
    test_args = ["moban", "-V"]
    with patch.object(sys, "argv", test_args):
        from moban.main import main

        main()


@patch("logging.basicConfig")
def test_warning_verbose(fake_config):
    fake_config.side_effect = [IOError("stop test")]
    test_args = ["moban", "-vvv"]
    with patch.object(sys, "argv", test_args):
        from moban.main import main

        try:
            main()
        except IOError:
            fake_config.assert_called_with(
                format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                level=10,
            )


@patch("logging.basicConfig")
def test_debug_five_verbose_option(fake_config, *_):
    fake_config.side_effect = [IOError("stop test")]
    test_args = ["moban", "-vvvvv"]
    with patch.object(sys, "argv", test_args):
        from moban.main import main

        try:
            main()
        except IOError:
            fake_config.assert_called_with(
                format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                level=10,
            )


@patch("moban.core.utils.verify_the_existence_of_directories", return_value=[])
def test_git_repo_example(_):
    test_args = [
        "moban",
        "-t",
        "git://github.com/moremoban/pypi-mobans.git!/templates/_version.py.jj2",
        "-c",
        "git://github.com/moremoban/pypi-mobans.git!/config/data.yml",
        "-o",
        "test_git_repo_example.py",
    ]
    with patch.object(sys, "argv", test_args):
        from moban.main import main

        main()
        with open("test_git_repo_example.py") as f:
            content = f.read()
            eq_(content, '__version__ = "0.1.1rc3"\n__author__ = "C.W."\n')
        os.unlink("test_git_repo_example.py")


@patch("moban.core.utils.verify_the_existence_of_directories", return_value=[])
def test_pypi_pkg_example(_):
    test_args = [
        "moban",
        "-t",
        "pypi://pypi-mobans-pkg/resources/templates/_version.py.jj2",
        "-c",
        "pypi://pypi-mobans-pkg/resources/config/data.yml",
        "-o",
        "test_pypi_pkg_example.py",
    ]
    with patch.object(sys, "argv", test_args):
        from moban.main import main

        main()
        with open("test_pypi_pkg_example.py") as f:
            content = f.read()
            eq_(content, '__version__ = "0.1.1rc3"\n__author__ = "C.W."\n')
        os.unlink("test_pypi_pkg_example.py")


def test_add_extension():
    if sys.version_info[0] == 2:
        raise SkipTest("jinja2-python-version does not support python 2")
    test_commands = [
        [
            "moban",
            "-t",
            "{{ python_version }}",
            "-e",
            "jinja2=jinja2_python_version.PythonVersionExtension",
        ],
        [
            "moban",
            "-t",
            "{{ python_version }}",
            "-e",
            "jj2=jinja2_python_version.PythonVersionExtension",
        ],
    ]
    for test_args in test_commands:
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            with open("moban.output") as f:
                content = f.read()
                eq_(
                    content,
                    "{}.{}".format(sys.version_info[0], sys.version_info[1]),
                )
            os.unlink("moban.output")


def test_stdin_input():
    if sys.platform == "win32":
        raise SkipTest("windows test fails with this pipe test 2")
    test_args = ["moban", "-d", "hello=world"]
    with patch.object(sys, "stdin", StringIO("{{hello}}")):
        with patch.object(sys, "argv", test_args):
            from moban.main import main

            main()
            with open("moban.output") as f:
                content = f.read()
                eq_(content, "world")
            os.unlink("moban.output")



import csv

from lml.plugin import PluginInfo

from moban import constants


@PluginInfo(constants.DATA_LOADER_EXTENSION, tags=["custom"])
def open_custom(file_name):
    with open(file_name, "r") as data_csv:
        csvreader = csv.reader(data_csv)
        rows = []
        for row in csvreader:
            rows.append(row)

        data = dict(zip(rows[0], rows[1]))
        return data



import fs.path
from nose.tools import eq_, raises

from moban.core.data_loader import load_data
from moban.plugins.yaml_loader import open_yaml


def test_simple_yaml():
    test_file = fs.path.join("tests", "fixtures", "simple.yaml")
    data = open_yaml(test_file)
    eq_(data, {"simple": "yaml"})


def test_inheritance_yaml():
    test_file = fs.path.join("tests", "fixtures", "child.yaml")
    data = load_data(fs.path.join("tests", "fixtures", "config"), test_file)
    eq_(data, {"key": "hello world", "pass": "ox"})


def test_exception():
    test_file = fs.path.join("tests", "fixtures", "orphan.yaml")
    data = load_data(fs.path.join("tests", "fixtures", "config"), test_file)
    eq_(len(data), 0)


@raises(IOError)
def test_exception_2():
    test_file = fs.path.join("tests", "fixtures", "dragon.yaml")
    load_data(fs.path.join("tests", "fixtures", "config"), test_file)


@raises(IOError)
def test_exception_3():
    test_file = fs.path.join("tests", "fixtures", "dragon.yaml")
    load_data(None, test_file)



class DirectoryNotFound(Exception):
    pass


class FileNotFound(Exception):
    pass


class NoThirdPartyEngine(Exception):
    pass


class MobanfileGrammarException(Exception):
    pass


class NoTemplate(Exception):
    pass


class IncorrectDataInput(Exception):
    pass


class GroupTargetNotFound(Exception):
    pass


class NoGitCommand(Exception):
    pass


class UnsupportedPyFS2Protocol(Exception):
    pass


class NoPermissionsNeeded(Exception):
    pass


class SingleHTTPURLConstraint(Exception):
    pass


class PassOn(Exception):
    """
    Raised when template engine cannot do anything with the given template.

    i.e. given a png image :/
    """

    pass
